#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

REGION_INFOS = [
    {
        "regionId": "cn-beijing",
        "regionName": [
            "China (Beijing)",
            "华北 2 （北京）",
        ],
    },
    {
        "regionId": "cn-hangzhou",
        "regionName": [
            "China (Hangzhou)",
            "华东 1 （杭州）",
        ],
    },
    # {
    #     "regionId": "cn-zhangjiakou",
    #     "regionName": [
    #         "China (Zhangjiakou)",
    #         "华北 3 （张家口）",
    #     ],
    # },
    {
        "regionId": "cn-shanghai",
        "regionName": [
            "China (Shanghai)",
            "华东 2 （上海）",
        ],
    },
    {
        "regionId": "cn-wulanchabu",
        "regionName": [
            "China (Wulanchabu)",
            "华北6（乌兰察布）",
        ],
    },
    {
        "regionId": "cn-shenzhen",
        "regionName": [
            "China (Shenzhen)",
            "华南 1 （深圳）",
        ],
    },
    {
        "regionId": "cn-hongkong",
        "regionName": [
            "China (Hongkong)",
            "中国香港",
        ],
    },
    {
        "regionId": "ap-southeast-1",
        "regionName": [
            "Singapore",
            "新加坡",
        ],
    },
    # {
    #     "regionId": "ap-southeast-2",
    #     "regionName": [
    #         "Australia (Sydney)",
    #         "澳大利亚（悉尼）",
    #     ],
    # },
    {
        "regionId": "ap-southeast-3",
        "regionName": [
            "Malaysia (Kuala Lumpur)",
            "马来西亚（吉隆坡）",
        ],
    },
    {
        "regionId": "ap-southeast-5",
        "regionName": [
            "Indonesia (Jakarta)",
            "印度尼西亚（雅加达）",
        ],
    },
    # {
    #     "regionId": "ap-south-1",
    #     "regionName": [
    #         "India (Mumbai)",
    #         "印度（孟买）",
    #     ],
    # },
    # {
    #     "regionId": "us-east-1",
    #     "regionName": [
    #         "US (Virginia)",
    #         "美国东部 1 （维吉尼亚）",
    #     ],
    # },
    # {
    #     "regionId": "us-west-1",
    #     "regionName": [
    #         "US (Silicon Valley)",
    #         "美国西部 1 （硅谷）",
    #     ],
    # },
    {
        "regionId": "eu-central-1",
        "regionName": [
            "Germany (Frankfurt)",
            "德国（法兰克福）",
        ],
    },
    # {
    #     "regionId": "me-east-1",
    #     "regionName": [
    #         "UAE (Dubai)",
    #         "阿联酋（迪拜）",
    #     ],
    # },
]
